/*
 * fsindex, create a file system index in "ls -l" format, including MD5 sums
 * Copyright (C) 2000 Frank van Maarseveen
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Written by Frank van Maarseveen <f.vanmaarseveen@inter.nl.net>
 */

/******************************************************************************
This program prints an index of all filesystem objects. The output
is intended for tracking changes in the file system and for searching potential
security holes. The output format consists
of one line per file system object. Every line contains 7 fields. The tab
character is used as field separator. For every field an embedded tab is
replaced by \t, linefeed by \n and backslashses are doubled.

Field	contents
0	mode of object (ls -l style: drwxr-xr-x).
1	link count
2	user id
3	group id
4	modified time. The epoch when not a directory or regular file.
5	pathname, relative to directory argument.
6	MD5 sum for plain files, symlink contents for symlinks, rdev for
	device inodes, otherwise 0
******************************************************************************/

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <fcntl.h>
#include <unistd.h>
#include <time.h>
#include <pwd.h>
#include <grp.h>

#include "md5.h"

static char	*myname = "fsindex";
static char	*usage[] = {
	"Usage: %s [options...] dir... [options...]",
	"",
	"Options are:",
	"    -md5          generate MD5 checksum for regular files",
	"    -xdev         do not traverse mount points",
	"    -exclude arg  Exclude the pathname argument from the traversal.",
	"    -depth        descend directory before handling directory itself",
	"    -localtime    use the local time zone instead of UTC",
	"    -insecure     find potential security problems in the file system:",
	"                  set[gu]id programs and writable objects",
	"    -name         print names of user/group id instead of their number",
	"",
	"    directory arguments and pathnames to be excluded should either be",
	"    both relative to the current working directory or both rooted.",
	"",
	NULL
};
static	int	opt_md5;		/* include MD5 checksum in output */
static	int	opt_xdev;		/* do not traverse mount points */
static	int	opt_excludec;		/* # of -exclude options */
static	int	opt_depth;		/* scan contents of directory before directory itself */
static	int	opt_localtime;
static	int	opt_insecure;
static	int	opt_name;
static	char	**opt_excludev;		/* vector of pathnames to exclude from output */
static	char	*cwd;			/* rooted pathname to initial working directory */
static	char	*path;			/* pathname being scanned, rooted or relative to cwd */
static	dev_t	current_device;
static	int	exit_status;

	int	main( int argc, char **argv );
static	void	scan( void );
static	void	process_object( char *ent, char *newpath, struct stat *st );
static	void	doit( char *ent, char *newpath, struct stat *st );
static	void	fingerprint( char *ent, char *newpath, struct stat *st );
static	char	*mode2str( mode_t mode );
static	char	*uid2str( uid_t uid );
static	char	*gid2str( gid_t gid );
static	char	*time2str( time_t t );
static	char	*md5( char *ent );
static	void	printname( char *s );
static	char	*connect_path( char *p1, char *p2 );
static	char	*entryof( char *p1, char *p2 );
static	void	print_usage( void );
static	int	is_excluded( char *s );
static	void	warn( char *fmt, ... );
static	void	error( char *fmt, ... );
static	void	fatal( char *fmt, ... );
static	void	*alc( int size );
static	char	*alcstr( char *s );
static	char	*alcsymlink( char *ent, int size );

int main( int argc, char **argv )
{
	int		i;
	int		dirc;			/* # of directory arguments */
	char		**dirv;			/* directory argument vector, not NULL terminated */
	struct stat	st;

	dirc = 0;
	dirv = alc( argc * sizeof(char *) );
	opt_excludev = alc( argc * sizeof(char *) );
	for (i = 1; i < argc; ++i) {
		if (strcmp( argv[i], "-md5" ) == 0) {
			opt_md5 = 1;
		} else if (strcmp( argv[i], "-xdev" ) == 0) {
			opt_xdev = 1;
		} else if (strcmp( argv[i], "-exclude" ) == 0) {
			if (++i >= argc) {
				fatal( "-exclude requires an argument\n" );
			}
			opt_excludev[opt_excludec++] = argv[i];
		} else if (strcmp( argv[i], "-depth" ) == 0) {
			opt_depth = 1;
		} else if (strcmp( argv[i], "-localtime" ) == 0) {
			opt_localtime = 1;
		} else if (strcmp( argv[i], "-insecure" ) == 0) {
			opt_insecure = 1;
		} else if (strcmp( argv[i], "-name" ) == 0) {
			opt_name = 1;
		} else if (argv[i][0] == '-') {
			print_usage();
			exit( 1 );
		} else {
			dirv[dirc++] = argv[i];
		}
	}
	if (!dirc) {
		dirv[dirc++] = ".";
	}
	for (i = PATH_MAX + 2; ; i *= 2) {
		cwd = alc( i );
		memset( cwd, 0, i );
		if (getcwd( cwd, i - 2 )) {
			break;
		}
		if (errno != ERANGE) {
			fatal( "getcwd: %s\n", strerror( errno ) );
		}
		free( cwd );
	}
	path = ".";
	for (i = 0; i < dirc; ++i) {
		if (lstat( dirv[i], &st )) {
			error( "%s: %s\n", dirv[i], strerror( errno ) );
			continue;
		}
		current_device = st.st_dev;
		process_object( dirv[i], dirv[i], &st );
	}
	free( dirv );
	free( opt_excludev );
	free( cwd );
	fflush( stdout );
	if (ferror( stdout) ) {
		fatal( "standard output: %s\n", strerror( errno ) );
	}
	return exit_status;
}

static void scan( void )
{
	DIR		*dir;
	struct dirent	*dirent;
	char		*ent;
	char		*newpath;
	struct stat	st;

	if ((dir = opendir( "." )) == NULL) {
		warn( "%s: %s, directory contents ignored (warning)\n", path, strerror( errno ) );
		return;
	}
	while ((dirent = readdir( dir ))) {
		ent = dirent->d_name;
		if (strcmp( ent, "." ) == 0 || strcmp( ent, ".." ) == 0) {
			continue;
		}
		if (lstat( ent, &st )) {
			error( "lstat %s (in %s): %s, consistency error\n", ent, path, strerror( errno ) );
			continue;
		}
		newpath = connect_path( path, ent );
		process_object( ent, newpath, &st );
		free( newpath );
	}
	if (closedir( dir )) {
		error( "%s: %s\n", path, strerror( errno ) );
	}
}

/*
 * Both `ent' and
 * `newpath' should point to the same object. When directories are
 * descended then the latter will be used for prefixing the names of
 * file system objects inside. Actually, `ent' is a shortcut for the
 * full path in `newpath' and is used for readlink, chdir, taking an
 * MD5, etc for saving pathname lookups in the kernel.
 *
 * `st' should be filled in by the caller, avoiding many lstat() calls
 * for the same object.
 *
 * This function uses chdir(): the current working directory for all
 * doit() and downward calls is the directory holding the object so `ent'
 * can be used for open(). When the object is a directory then the current working
 * directory is set to the directory, i.e. use "." in operations. The initial
 * object is the specified arument on the commandline so `ent' may actually
 * be a pathname.
 */
static void process_object( char *ent, char *newpath, struct stat *st )
{
	char		*oldpath;
	char		*back;

	if (is_excluded( newpath )) {
		return;
	}
	switch (st->st_mode & S_IFMT) {
	case S_IFDIR:
		if (opt_xdev && st->st_dev != current_device) {
			break;
		}
		if (chdir( ent )) {
			if (!opt_insecure) {
				error( "chdir %s (in %s): %s\n", ent, path, strerror( errno ) );
			}
			break;
		}
		oldpath = path;
		path = newpath;
		if (opt_depth) {
			scan();
			doit( ent, newpath, st );
		} else {
			doit( ent, newpath, st );
			scan();
		}
		path = oldpath;
		back = connect_path( cwd, path );
		if (chdir( back )) {
			fatal( "chdir %s: %s\n", back, strerror( errno ) );
		}
		free( back );
		break;
	default:
		doit( ent, newpath, st );
		break;
	}
	if (ferror( stdout) ) {
		fatal( "standard output: %s\n", strerror( errno ) );
	}
}

static void doit( char *ent, char *newpath, struct stat *st )
{
	uid_t	uid;

	switch (st->st_mode & S_IFMT) {
#ifdef S_IFDOOR
	case S_IFDOOR:
		warn( "%s (in %s): is a door, ignored (warning)\n", ent, path );
		return;
#endif
	default:
		break;
	}
	uid = getuid();
	if (opt_insecure) {
		switch (st->st_mode & S_IFMT) {
		case S_IFLNK:
			if (st->st_uid == uid) {
				fingerprint( ent, newpath, st );
			}
			break;
		case S_IFSOCK:
		case S_IFIFO:
		case S_IFBLK:
		case S_IFCHR:
			if (st->st_uid == uid || access( ent, W_OK ) == 0) {
				fingerprint( ent, newpath, st );
			}
			break;
		case S_IFREG:
			if (st->st_uid == uid || access( ent, W_OK ) == 0 || ((st->st_mode & (S_ISUID | S_ISGID)) && access( ent, X_OK ) == 0)) {
				fingerprint( ent, newpath, st );
			}
			break;
		case S_IFDIR:
			if (st->st_uid == uid || (!(st->st_mode & S_ISVTX) && access( ent, W_OK ) == 0)) {
				fingerprint( ent, newpath, st );
			}
			break;
		default:
			fatal( "%s (in %s): unknown file system object (fmt=0%o)\n", ent, path, st->st_mode & S_IFMT );
			break;
		}
	} else {
		fingerprint( ent, newpath, st );
	}
}

static void fingerprint( char *ent, char *newpath, struct stat *st )
{
	char		*s;
	int		i;
	struct stat	st2;
	unsigned int	nlink;

	switch (st->st_mode & S_IFMT) {
	case S_IFSOCK:
	case S_IFIFO:
		printf( "%s\t%u\t%s\t%s\t%s\t", mode2str( st->st_mode ), st->st_nlink, uid2str( st->st_uid ), gid2str( st->st_gid ), time2str( 0 ) );
		printname( newpath );
		printf( "\t%s\n", "0" );
		break;
	case S_IFLNK:
		if ((s = alcsymlink( ent, st->st_size )) == NULL) {
			break;
		}
		printf( "%s\t%u\t%s\t%s\t%s\t", mode2str( st->st_mode ), st->st_nlink, uid2str( st->st_uid ), gid2str( st->st_gid ), time2str( 0 ) );
		printname( newpath );
		printf( "\t" );
		printname( s );
		printf( "\n" );
		free( s );
		break;
	case S_IFREG:
		printf( "%s\t%u\t%s\t%s\t%s\t", mode2str( st->st_mode ), st->st_nlink, uid2str( st->st_uid ), gid2str( st->st_gid ), time2str( st->st_mtime ) );
		printname( newpath );
		printf( "\t%s\n", opt_md5 ? md5( ent ) : "0" );
		break;
	case S_IFBLK:
	case S_IFCHR:
		printf( "%s\t%u\t%s\t%s\t%s\t", mode2str( st->st_mode ), st->st_nlink, uid2str( st->st_uid ), gid2str( st->st_gid ), time2str( 0 ) );
		printname( newpath );
#ifdef __linux__
		printf( "\t%08qx\n", st->st_rdev );
#else
		printf( "\t%08x\n", st->st_rdev );
#endif
		break;
	case S_IFDIR:
		/*
		 * When a directory is excluded and it exists and its's an entry of `newpath' then
		 * the link count of `newpath' needs a correction because of the ".." hard link inside the
		 * excluded directory.
		 */
		nlink = st->st_nlink;
		for (i = 0; i < opt_excludec; ++i) {
			if ((s = entryof( newpath, opt_excludev[i] ))) {
				nlink -= lstat( s, &st2 ) == 0 && ((st2.st_mode & S_IFMT) == S_IFDIR);
			}
		}
		printf( "%s\t%u\t%s\t%s\t%s\t", mode2str( st->st_mode ), nlink, uid2str( st->st_uid ), gid2str( st->st_gid ), time2str( st->st_mtime ) );
		printname( newpath );
		printf( "\t%s\n", "0" );
		break;
	default:
		fatal( "unknown file system object (fmt=0%o)\n", st->st_mode & S_IFMT );
		break;
	}
}

static char *mode2str( mode_t mode )
{
	static char	buf[11];	/* drwxr-xr-x */

	switch (mode & S_IFMT) {
	case S_IFSOCK:	buf[0] = 's';	break;
	case S_IFLNK:	buf[0] = 'l';	break;
	case S_IFREG:	buf[0] = '-';	break;
	case S_IFBLK:	buf[0] = 'b';	break;
	case S_IFDIR:	buf[0] = 'd';	break;
	case S_IFCHR:	buf[0] = 'c';	break;
	case S_IFIFO:	buf[0] = 'p';	break;
	default:
		fatal( "unknown file system object (fmt=0%o)\n", mode & S_IFMT );
		break;
	}
	buf[1] = mode &  S_IREAD	? 'r' : '-';
	buf[2] = mode &  S_IWRITE	? 'w' : '-';
	buf[3] = mode &  S_IEXEC	? 'x' : '-';
	buf[4] = mode & (S_IREAD  >> 3)	? 'r' : '-';
	buf[5] = mode & (S_IWRITE >> 3)	? 'w' : '-';
	buf[6] = mode & (S_IEXEC  >> 3)	? 'x' : '-';
	buf[7] = mode & (S_IREAD  >> 6)	? 'r' : '-';
	buf[8] = mode & (S_IWRITE >> 6)	? 'w' : '-';
	buf[9] = mode & (S_IEXEC  >> 6)	? 'x' : '-';
	if (mode & S_ISUID) {
		buf[3] = buf[3] == 'x' ? 's' : 'S';
	}
	if (mode & S_ISGID) {
		buf[6] = buf[6] == 'x' ? 's' : 'S';
	}
	if (mode & S_ISVTX) {
		buf[9] = buf[9] == 'x' ? 't' : 'T';
	}
	return buf;
}

static char *uid2str( uid_t uid )
{
	struct passwd	*pw;
	static char	*buf;
	static int	size;

	if (!size) {
		size = 5 * sizeof( uid_t ) + 1;		/* a (worst case 16 bits wide) char is expressable in 5 decimal digits */
		buf = alc( size );
	}
	if (opt_name) {
		if ((pw = getpwuid( uid ))) {
			if (strlen( pw->pw_name ) >= size) {
				free( buf );
				size = 2 * strlen( pw->pw_name ) + 1;
				buf = alc( size );
			}
			strcpy( buf, pw->pw_name );
		} else {
			if (errno == ENOMEM) {
				fatal( "getpwuid: %s", strerror( errno ) );
			}
			sprintf( buf, "%u", uid );
		}
	} else {
		sprintf( buf, "%u", uid );
	}
	return buf;
}

static char *gid2str( gid_t gid )
{
	struct group	*gr;
	static char	*buf;
	static int	size;

	if (!size) {
		size = 5 * sizeof( gid_t ) + 1;		/* a (worst case 16 bits wide) char is expressable in 5 decimal digits */
		buf = alc( size );
	}
	if (opt_name) {
		if ((gr = getgrgid( gid ))) {
			if (strlen( gr->gr_name ) >= size) {
				free( buf );
				size = 2 * strlen( gr->gr_name ) + 1;
				buf = alc( size );
			}
			strcpy( buf, gr->gr_name );
		} else {
			if (errno == ENOMEM) {
				fatal( "getgrgid: %s", strerror( errno ) );
			}
			sprintf( buf, "%u", gid );
		}
	} else {
		sprintf( buf, "%u", gid );
	}
	return buf;
}

static char *
time2str( time_t t )
{
	struct tm	*tm;
	static char	buf[32];		/* 18 should do: YYYYmmdd HH:MM:SS */

	tm = opt_localtime ? localtime( &t ) : gmtime( &t );
	strftime( buf, sizeof(buf) - 1, "%Y%m%d %H:%M:%S", tm );		/* big endian notation, good for sorting */
	buf[sizeof(buf) - 1] = 0;
	return buf;
}

/*
 * In order to make sure all output lines contain the same number of fields
 * this function returns a nonempty string on an error which is clearly
 * not a valid MD5 message digest.
 * The return value points to static memory and is overwritten in the next call.
 */
static char *md5( char *ent )
{
	unsigned char	md5sum[16];
	static char	md5str[33];
	int		i, d;

	if ((d = open( ent, O_RDONLY, 0 )) < 0) {
		error( "cannot open %s (in %s): %s\n", ent, path, strerror( errno ) );
		return "-1";
	}
	if (md5_stream( d, md5sum ) || close( d )) {
		error( "%s (in %s): %s\n", ent, path, strerror( errno ) );
		return "-1";
	}
	for (i = 0; i < sizeof(md5sum); ++i) {
		md5str[2 * i    ] = "0123456789abcdef"[(md5sum[i] >> 4) & 0xf];
		md5str[2 * i + 1] = "0123456789abcdef"[(md5sum[i]     ) & 0xf];
	}
	md5str[sizeof(md5str) - 1] = 0;
	return md5str;
}

/*
 * \t and \n are special so they have to be escaped. The escape character '\\'
 * is therefore special as well. The function below is intended for printing strings
 * which might contain special characters such as filenames and link contents. It's
 * bad practice to use these characters in file names but we should be able to handle
 * them (not doing this would be bad practice as well)
 */
static void printname( char *s )
{
	char	c;

	while ((c = *s++)) {
		switch (c) {
		case '\t':
			putchar( '\\' );
			putchar( 't' );
			break;
		case '\n':
			putchar( '\\' );
			putchar( 'n' );
			break;
		case '\\':
			putchar( '\\' );
			putchar( '\\' );
			break;
		default:
			putchar( c );
			break;
		}
	}
}

/*
 * Connect the two pathnames and return the resulting pathname. The
 * return value must be freed through free() after use.
 * Examples:
 *	p1	p2	result
 *	----------------------
 *	/a	b	/a/b
 *	/	b	/b
 *	/a	/b	/b
 */
static char *connect_path( char *p1, char *p2 )
{
	char	*result;

	if (*p2 == '/') {
		result = alcstr( p2 );
	} else if (strcmp( p1, "/" ) == 0) {
		result = alc( strlen( p1 ) + strlen( p2 ) + 1 );
		sprintf( result, "%s%s", p1, p2 );
	} else {
		result = alc( strlen( p1 ) + 1 + strlen( p2 ) + 1 );
		sprintf( result, "%s/%s", p1, p2 );
	}
	return result;
}

/*
 * Return the basename part of the second pathname if the second pathname
 * is an entry of the first one. Conform to the usage both pathnames
 * are rooted or relative but not mixed. Examples:
 *
 *	p1	p2	result
 *	----------------------
 *	/a	/a/b	b
 *	/a	/a/b/c	NULL
 *	a	a/b	b
 *	a	a/b/c	NULL
 *	/	/a	a
 */
static char *entryof( char *p1, char *p2 )
{
	int	n1, n2, isroot;
	char	*ent;

	ent = NULL;
	isroot = strcmp( p1, "/" ) == 0;
	n1 = strlen( p1 );
	n2 = strlen( p2 );
	if (n2 > n1 && memcmp( p1, p2, n1 ) == 0 && (isroot || p2[n1] == '/')) {
		ent = p2 + n1 + !isroot;
		if (strchr( ent, '/' )) {
			ent = NULL;
		}
	}
	return ent;
}

static void print_usage( void )
{
	char	**p;

	for (p = usage; *p; ++p) {
		printf( *p, myname );
		putchar( '\n' );
	}
}

static int is_excluded( char *s )
{
	int	i;

	for (i = 0; i < opt_excludec; ++i) {
		if (strcmp( opt_excludev[i], s ) == 0) {
			break;
		}
	}
	return i < opt_excludec;
}

static void warn( char *fmt, ... )
{
	va_list	ap;

	fprintf( stderr, "%s: ", myname );
	va_start(ap,fmt);
	vfprintf( stderr, fmt, ap );
	va_end(ap);
}

static void error( char *fmt, ... )
{
	va_list	ap;

	fprintf( stderr, "%s: ", myname );
	va_start(ap,fmt);
	vfprintf( stderr, fmt, ap );
	va_end(ap);
	exit_status = 1;
}

static void fatal( char *fmt, ... )
{
	va_list	ap;

	fprintf( stderr, "%s: ", myname );
	va_start(ap,fmt);
	vfprintf( stderr, fmt, ap );
	va_end(ap);
	exit( 2 );
}

static void *alc( int size )
{
	void	*rv;

	rv = malloc( size );
	if (rv == NULL) {
		fatal( "not enough memory.\n" );
	}
	return rv;
}

static char *alcstr( char *s )
{
	char	*copy;

	copy = alc( strlen( s ) + 1 );
	strcpy( copy, s );
	return copy;
}

static char *alcsymlink( char *ent, int size )
{
	char	*s;
	int	len;

	s = alc( size + 1 );
	len = readlink( ent, s, size + 1 );
	if (len < 0) {
		error( "readlink %s (in %s): %s\n", ent, path, strerror( errno ) );
		free( s );
		s = NULL;
	} else if (len > size) {
		error( "readlink %s (in %s): consistency error, link contents don't fit\n", ent, path );
		free( s );
		s = NULL;
	} else {
		s[len] = 0;
	}
	return s;
}
